"""
The implementation is largely based on the following reference:
Van der Laan MJ, Rose S. Targeted learning: causal inference for observational and experimental data. Springer; New York: 2011. Specifically, Chapter 8 for the ATT TMLE.
But slightly modified for simpler implementation, following advice from: https://stats.stackexchange.com/questions/520472/can-targeted-maximum-likelihood-estimation-find-the-average-treatment-effect-on/534018#534018
"""

import warnings

import numpy as np
from scipy.special import expit, logit
from statsmodels.genmod.families import Binomial
from statsmodels.genmod.generalized_linear_model import GLM


def compute_estimates_att(A, Y, ps, Y0_hat, Y1_hat, Yhat):
    """
    Compute updated outcome estimates for ATT using a one-step TMLE targeting step.

    For ATT, the clever covariate is defined as:
      H(A,W) = 1{A=1}/P(A=1) - 1{A=0}*ps/(P(A=1)*(1-ps)),
    where P(A=1) is the empirical probability of treatment.

    Parameters:
    -----------
    A: array-like
         Treatment assignment (0 or 1)
    Y: array-like
         Binary outcome
    ps: array-like
         Propensity score P(A=1|X)
    Y0_hat: array-like
         Initial outcome prediction for control group P(Y=1|A=0,X)
    Y1_hat: array-like
         Initial outcome prediction for treatment group P(Y=1|A=1,X)
    Yhat: array-like
         Combined outcome prediction = A*Y1_hat + (1-A)*Y0_hat

    Returns:
    --------
    tuple: (Q_star_1, Q_star_0)
         Updated outcome predictions for treated and control groups.
    """
    # Empirical probability of treatment
    p_treated = np.mean(A == 1)

    # Estimate the fluctuation parameter epsilon using a logistic regression:
    epsilon = estimate_fluctuation_parameter_att(A, Y, ps, Yhat)

    # Update outcome predictions:
    # For treated group: H = 1/p_treated, so update is: logit(Y1_hat) + epsilon*(1/p_treated)
    # For control group: H = -ps/(p_treated*(1-ps)), so update is: logit(Y0_hat) - epsilon*(ps/(p_treated*(1-ps)))
    Q_star_1 = expit(logit(Y1_hat) + epsilon * (1.0 / p_treated))
    Q_star_0 = expit(logit(Y0_hat) - epsilon * (ps / (p_treated * (1 - ps))))

    return Q_star_1, Q_star_0


def estimate_fluctuation_parameter_att(A, Y, ps, Yhat) -> float:
    """
    Estimate the fluctuation parameter epsilon for the ATT TMLE via logistic regression.

    The model is:
         logit(Q*(A,W)) = logit(Yhat) + epsilon * H,
    and is fitted with no intercept.

    Parameters:
    -----------
    A: array-like
         Treatment assignment (0 or 1)
    Y: array-like
         Binary outcome
    ps: array-like
         Propensity score P(A=1|X)
    Yhat: array-like
         Combined outcome prediction (A*Y1_hat + (1-A)*Y0_hat)

    Returns:
    --------
    float: Estimated fluctuation parameter epsilon.
    """
    p_treated = np.mean(A == 1)
    # Define the clever covariate H for each observation:
    # For treated: 1/p_treated, for controls: -ps/(p_treated*(1-ps))
    H = (A / p_treated) - ((1 - A) * ps / (p_treated * (1 - ps)))
    # Check for extreme values in H
    if np.any(np.abs(H) > 100):
        warnings.warn(
            "Extreme values detected in clever covariate H. "
            "This may indicate issues with propensity scores near 0 or 1.",
            RuntimeWarning,
        )
    # Reshape H for statsmodels (needs a 2D array)
    H_2d = H.reshape(-1, 1)
    offset = logit(Yhat)

    # Fit logistic regression with no intercept (i.e. model: Y ~ -1 + H, with offset)
    model = GLM(Y, H_2d, family=Binomial(), offset=offset)
    results = model.fit()

    epsilon = results.params[0]
    return epsilon


def compute_tmle_att(A, Y, ps, Y0_hat, Y1_hat, Yhat):
    """
    Estimate the Average Treatment Effect on the Treated (ATT) using TMLE.

    The TMLE ATT estimator is given by:
      Ψ* = (1/n) * sum [ ps(W_i)/P(A=1) * ( Q*_n(1,W_i) - Q*_n(0,W_i) ) ],
    where Q*_n(1,W) and Q*_n(0,W) are the updated outcome predictions.

    Parameters:
    -----------
    A: array-like
         Treatment assignment (0 or 1)
    Y: array-like
         Binary outcome
    ps: array-like
         Propensity score P(A=1|X)
    Y0_hat: array-like
         Initial outcome prediction for controls
    Y1_hat: array-like
         Initial outcome prediction for treated
    Yhat: array-like
         Combined outcome prediction = A*Y1_hat + (1-A)*Y0_hat

    Returns:
    --------
    float: ATT estimate.
    """
    Q_star_1, Q_star_0 = compute_estimates_att(A, Y, ps, Y0_hat, Y1_hat, Yhat)
    p_treated = np.mean(A == 1)

    # Compute plugin ATT estimator: average of (Q*_1 - Q*_0) over treated units reweighted by ps/P(A=1)
    psi = np.mean((ps / p_treated) * (Q_star_1 - Q_star_0))
    return psi



import pandas as pd

from CausalEstimate.estimators.base import BaseEstimator
from CausalEstimate.estimators.functional.ipw import (
    compute_ipw_ate,
    compute_ipw_ate_stabilized,
    compute_ipw_att,
    compute_ipw_risk_ratio,
    compute_ipw_risk_ratio_treated,
)


class IPW(BaseEstimator):
    def __init__(
        self,
        effect_type="ATE",
        treatment_col="treatment",
        outcome_col="outcome",
        ps_col="ps",
        **kwargs,
    ):
        super().__init__(effect_type=effect_type, **kwargs)
        self.treatment_col = treatment_col
        self.outcome_col = outcome_col
        self.ps_col = ps_col
        self.kwargs = kwargs

    def _compute_effect(self, df: pd.DataFrame) -> float:
        """
        Compute the effect using the functional IPW.
        Available effect types: ATE, ATT, RR, RRT
        """
        A = df[self.treatment_col]
        Y = df[self.outcome_col]
        ps = df[self.ps_col]
        if self.effect_type in ["ATE", "ARR"]:
            if self.kwargs.get("stabilized", False):
                return compute_ipw_ate_stabilized(A, Y, ps)
            else:
                return compute_ipw_ate(A, Y, ps)
        elif self.effect_type in ["ATT"]:
            return compute_ipw_att(A, Y, ps)
        elif self.effect_type == "RR":
            return compute_ipw_risk_ratio(A, Y, ps)
        elif self.effect_type == "RRT":
            return compute_ipw_risk_ratio_treated(A, Y, ps)
        else:
            raise ValueError(f"Effect type '{self.effect_type}' is not supported.")



# CausalEstimate/estimators/base.py
import warnings
from abc import ABC, abstractmethod
from typing import Literal

import pandas as pd


class BaseEstimator(ABC):
    VALID_EFFECT_TYPES = {"ATE", "ARR", "ATT", "ATC", "RR", "RRT"}  #
    PS_EXTREME_THRESHOLD = 0.01  # Configurable threshold for extreme PS warning
    MIN_GROUP_SIZE = 10  # Configurable minimum group size

    def __init__(
        self,
        effect_type: Literal[
            "ATE", "ARR", "ATT", "ATC", "RR", "RRT"
        ] = "ATE",  # if binary it is equivalent
        treatment_col: str = "treatment",
        outcome_col: str = "outcome",
        ps_col: str = "ps",
        **kwargs,
    ):
        """
        Base class for all estimators.
        - effect_type: e.g. "ATE", "ATT", ... (ARR absolute risk reduction is equivalent to ATE if binary)
        - treatment_col, outcome_col, ps_col: universal column names
        - kwargs: any additional method-specific settings or toggles
        """
        self.effect_type = effect_type
        self.treatment_col = treatment_col
        self.outcome_col = outcome_col
        self.ps_col = ps_col
        self.kwargs = kwargs  # if child classes want extra toggles

    def compute_effect(self, df: pd.DataFrame) -> float:
        """
        Compute the causal effect from the given dataframe.
        The columns to use are already known from the constructor.
        This method is to be implemented by child classes.
        """
        self._validate_input_df(df)
        return self._compute_effect(df)

    @abstractmethod
    def _compute_effect(self, df: pd.DataFrame) -> float:
        """
        Concrete implementation to be provided by child classes.
        Input df is guaranteed to be validated.
        """
        pass

    def _validate_input_df(self, df: pd.DataFrame) -> None:
        """
        Main validation method that orchestrates all validation checks.

        Raises:
            TypeError: If input is not a pandas DataFrame
            ValueError: If any validation fails
        Warns:
            RuntimeWarning: For concerning but non-fatal issues
        """
        self._validate_df_type(df)
        self._validate_columns_exist(df)

        # Extract columns once for efficiency
        treatment = df[self.treatment_col]
        ps = df[self.ps_col]

        outcome = df[self.outcome_col]

        self._validate_propensity_scores(ps)
        self._validate_treatment_values(treatment)
        self._validate_group_sizes(treatment)
        self._validate_missing_values(df)

        if self.effect_type in {
            "RR",
            "RRT",
            "ARR",
        }:  # should be binary for ratio-based estimands
            self._validate_binary_outcome(outcome)

    def _validate_df_type(self, df: pd.DataFrame) -> None:
        """Validate input is a pandas DataFrame."""
        if not isinstance(df, pd.DataFrame):
            raise TypeError("Input must be a pandas DataFrame")

    def _validate_columns_exist(self, df: pd.DataFrame) -> None:
        """Check all required columns are present."""
        required_cols = {self.treatment_col, self.outcome_col, self.ps_col}
        missing_cols = required_cols - set(df.columns)
        if missing_cols:
            raise ValueError(f"Missing required columns: {missing_cols}")

    def _validate_propensity_scores(self, ps: pd.Series) -> None:
        """Validate propensity scores range and check for extreme values."""
        # Check basic range
        if not ps.between(0, 1).all():
            raise ValueError("Propensity scores must be between 0 and 1")

        # Check for extreme values
        extreme_ps = (ps < self.PS_EXTREME_THRESHOLD) | (
            ps > (1 - self.PS_EXTREME_THRESHOLD)
        )
        if extreme_ps.any():
            n_extreme = extreme_ps.sum()
            warnings.warn(
                f"Found {n_extreme} ({n_extreme/len(ps):.1%}) propensity scores "
                f"outside [{self.PS_EXTREME_THRESHOLD}, {1-self.PS_EXTREME_THRESHOLD}]",
                RuntimeWarning,
            )

    def _validate_treatment_values(self, treatment: pd.Series) -> None:
        """Validate treatment is binary."""
        unique_treatments = set(treatment.unique())
        if not unique_treatments.issubset({0, 1}):
            raise ValueError(
                f"Treatment must be binary (0 or 1), found values: {unique_treatments}"
            )

    def _validate_group_sizes(self, treatment: pd.Series) -> None:
        """Check for sufficient sample sizes in treatment groups."""
        n_treated = (treatment == 1).sum()
        n_control = (treatment == 0).sum()

        if n_treated < self.MIN_GROUP_SIZE or n_control < self.MIN_GROUP_SIZE:
            warnings.warn(
                f"Small group sizes detected: treated={n_treated}, control={n_control}. "
                f"Minimum recommended size is {self.MIN_GROUP_SIZE}.",
                RuntimeWarning,
            )

    def _validate_missing_values(self, df: pd.DataFrame) -> None:
        """Check for missing values in required columns."""
        required_cols = {self.treatment_col, self.outcome_col, self.ps_col}
        for col in required_cols:
            n_missing = df[col].isna().sum()
            if n_missing > 0:
                raise ValueError(f"Found {n_missing} missing values in column '{col}'")

    def _validate_binary_outcome(self, outcome: pd.Series) -> None:
        """Validate outcome is binary for ratio-based estimands."""
        if not outcome.between(0, 1).all():
            raise ValueError(
                f"Outcome must be binary for {self.effect_type}, "
                f"found values outside [0,1]"
            )



from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def plot_outcome_proba_dist(
    df: pd.DataFrame,
    outcome_proba_col: str,
    treatment_col: str,
    xlabel: str = "Predicted Outcome Probability",
    title: str = "Outcome Probability Distribution",
    bin_edges: np.ndarray = None,
    normalize: bool = False,
    fig: plt.Figure = None,
    ax: plt.Axes = None,
    figsize: tuple = (10, 6),
):
    """
    Plot a predicted-outcome probability distribution for treatment vs. control groups.
    E.g., if 'outcome_proba_col' stores model-predicted probabilities.
    """
    return plot_hist_by_groups(
        df=df,
        value_col=outcome_proba_col,
        group_col=treatment_col,
        group_values=(0, 1),
        group_labels=("Control", "Treatment"),
        bin_edges=bin_edges,
        normalize=normalize,
        xlabel=xlabel,
        title=title,
        fig=fig,
        ax=ax,
        figsize=figsize,
    )


def plot_propensity_score_dist(
    df: pd.DataFrame,
    ps_col: str,
    treatment_col: str,
    xlabel: str = "Propensity Score",
    title: str = "Propensity Score Distribution",
    bin_edges: np.ndarray = None,
    normalize: bool = False,
    fig: plt.Figure = None,
    ax: plt.Axes = None,
    figsize: tuple = (10, 6),
):
    """
    Plot a propensity score distribution for treatment and control groups.
    """
    return plot_hist_by_groups(
        df=df,
        value_col=ps_col,
        group_col=treatment_col,
        group_values=(0, 1),
        group_labels=("Control", "Treatment"),
        bin_edges=bin_edges,
        normalize=normalize,
        xlabel=xlabel,
        title=title,
        fig=fig,
        ax=ax,
        figsize=figsize,
    )


def plot_hist_by_groups(
    df: pd.DataFrame,
    value_col: str,
    group_col: str,
    group_values=(0, 1),
    group_labels=("Group 0", "Group 1"),
    bin_edges=None,
    normalize: bool = False,
    xlabel: str = None,
    title: str = None,
    alpha: float = 0.5,
    colors=("b", "r"),
    fig: plt.Figure = None,
    ax: plt.Axes = None,
    figsize: tuple = (10, 6),
) -> Tuple[plt.Figure, plt.Axes]:
    """
    A generic helper that plots a histogram of 'value_col' for two groups
    defined by 'group_col', e.g. group_col=0 vs. group_col=1.

    Args:
        df (pd.DataFrame): DataFrame containing the data.
        value_col (str): The column whose distribution we want to plot.
        group_col (str): The column that indicates group membership.
        group_values (tuple): The two distinct values used to split the DataFrame.
        group_labels (tuple): Labels for legend (e.g. "Control", "Treatment").
        bin_edges (array): The bin edges for histogram. If None, defaults to 50 bins from 0..1
        normalize (bool): Whether to normalize the histogram (density=True).
        xlabel (str): X-axis label.
        title (str): Plot title.
        alpha (float): Transparency for the histogram overlay.
        colors (tuple): Colors for the two histograms.
        fig, ax: If provided, plot into them; otherwise create new figure/axes.
        figsize (tuple): Size of figure if we create a new one.

    Returns:
        (fig, ax)
    """
    # create or reuse figure/axes
    if fig is None and ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    elif ax is None:
        ax = fig.add_subplot(111)
    elif fig is None:
        raise ValueError("fig and ax cannot both be None")

    # default bins
    if bin_edges is None:
        bin_edges = np.linspace(0, 1, 51)  # 50 bins in [0,1]

    # group 0
    mask0 = df[group_col] == group_values[0]
    ax.hist(
        df.loc[mask0, value_col],
        bins=bin_edges,
        alpha=alpha,
        label=group_labels[0],
        color=colors[0],
        density=normalize,
    )

    # group 1
    mask1 = df[group_col] == group_values[1]
    ax.hist(
        df.loc[mask1, value_col],
        bins=bin_edges,
        alpha=alpha,
        label=group_labels[1],
        color=colors[1],
        density=normalize,
    )

    ax.set_xlabel(xlabel if xlabel else value_col)
    ax.set_ylabel("Count" if not normalize else "Density")
    if title:
        ax.set_title(title)
    ax.legend()

    return fig, ax



import unittest
from typing import List

import numpy as np
from scipy.special import expit

from CausalEstimate.simulation.binary_simulation import (
    compute_ATE_theoretical_from_data,
    compute_ATT_theoretical_from_data,
    compute_RR_theoretical_from_data,
    simulate_binary_data,
)
from CausalEstimate.utils.constants import (
    OUTCOME_COL,
    PID_COL,
    PROBAS_COL,
    PROBAS_T0_COL,
    PROBAS_T1_COL,
    PS_COL,
    TREATMENT_COL,
)


class TestEffectBase(unittest.TestCase):
    n: int = 10000
    alpha: List[float] = [0.1, 0.2, -0.3, 0]
    beta: List[float] = [0.5, 0.8, -0.6, 0.3, 0]
    seed: int = 42

    @classmethod
    def setUpClass(cls):
        # Simulate realistic data for testing
        rng = np.random.default_rng(cls.seed)
        # Covariates
        data = simulate_binary_data(
            cls.n, alpha=cls.alpha, beta=cls.beta, seed=cls.seed
        )

        # Predicted outcomes
        X = data[["X1", "X2"]].values
        A = data[TREATMENT_COL].values
        Y = data[OUTCOME_COL].values
        ps = expit(
            cls.alpha[0] + cls.alpha[1] * X[:, 0] + cls.alpha[2] * X[:, 1]
        ) + 0.01 * rng.normal(size=cls.n)
        Y1_hat = expit(
            cls.beta[0]
            + cls.beta[1] * 1
            + cls.beta[2] * X[:, 0]
            + cls.beta[3] * X[:, 1]
        ) + 0.01 * rng.normal(size=cls.n)
        Y0_hat = expit(
            cls.beta[0] + cls.beta[2] * X[:, 0] + cls.beta[3] * X[:, 1]
        ) + 0.01 * rng.normal(size=cls.n)
        Yhat = expit(
            cls.beta[0]
            + cls.beta[1] * A
            + cls.beta[2] * X[:, 0]
            + cls.beta[3] * X[:, 1]
        ) + 0.01 * rng.normal(size=cls.n)

        # clip
        eps = 1e-6
        Yhat = np.clip(Yhat, eps, 1 - eps)
        Y1_hat = np.clip(Y1_hat, eps, 1 - eps)
        Y0_hat = np.clip(Y0_hat, eps, 1 - eps)
        ps = np.clip(ps, eps, 1 - eps)

        cls.A = A
        cls.Y = Y
        cls.ps = ps
        cls.Y1_hat = Y1_hat
        cls.Y0_hat = Y0_hat
        cls.Yhat = Yhat

        true_ate = compute_ATE_theoretical_from_data(data, beta=cls.beta)
        true_att = compute_ATT_theoretical_from_data(data, beta=cls.beta)
        true_rr = compute_RR_theoretical_from_data(data, beta=cls.beta)
        cls.true_ate = true_ate
        cls.true_att = true_att
        cls.true_rr = true_rr

        # for classes that take dataframe as input
        cls.data = data
        cls.data[PID_COL] = np.arange(len(data))
        cls.data[TREATMENT_COL] = A
        cls.data[OUTCOME_COL] = Y
        cls.data[PS_COL] = ps
        cls.data[PROBAS_T1_COL] = Y1_hat
        cls.data[PROBAS_T0_COL] = Y0_hat
        cls.data[PROBAS_COL] = Yhat



import warnings

import numpy as np

from CausalEstimate.estimators.base import BaseEstimator
from CausalEstimate.estimators.functional.matching import compute_matching_ate
from CausalEstimate.matching.matching import match_eager, match_optimal
from CausalEstimate.utils.constants import OUTCOME_COL, PS_COL, TREATMENT_COL


class Matching(BaseEstimator):
    def __init__(
        self,
        effect_type="ATE",
        treatment_col=TREATMENT_COL,
        outcome_col=OUTCOME_COL,
        ps_col=PS_COL,
        match_optimal=True,
        **kwargs,
    ):
        super().__init__(effect_type=effect_type, **kwargs)
        self.treatment_col = treatment_col
        self.outcome_col = outcome_col
        self.ps_col = ps_col
        self.match_optimal = match_optimal
        self.kwargs = kwargs

    def _compute_effect(
        self,
        df,
    ) -> float:
        """
        Compute the effect using matching.
        Available effect types: ATE

        Effect computation for matched groups is different from the functional IPW.
        The matching itself influences the selected population, and thus the type of effect computed.
        E.g. when chosing ALL the trated and matching untreated to them, we get the ATT.
        However, when setting a caliper, we get a differnt population which is neither ATT nor ATE.[1]
        For ATE we can use full matching [2]

        [1] Greifer, Estimating Effects After Matching (https://cran.r-project.org/web/packages/MatchIt/vignettes/estimating-effects.html)
        [2] Stuart, et. al. "Using full matching to estimate causal effects in nonexperimental studies:
                examining the relationship between adolescent marijuana use and adult outcomes."
                Developmental psychology 44.2 (2008): 395.
        """
        Y = df[self.outcome_col]
        df = df.copy()  # Create a copy to avoid SettingWithCopyWarning

        # Add tiny random noise to propensity scores to break ties
        eps = 1e-10  # Small enough to not meaningfully affect matching
        df[self.ps_col] = df[self.ps_col] + np.random.uniform(-eps, eps, size=len(df))
        # Ensure PS stays in [0,1] range
        df[self.ps_col] = df[self.ps_col].clip(0, 1)

        df["index"] = range(
            len(df)
        )  # This ensures unique PIDs even with bootstrap samples

        if self.match_optimal:
            matched = match_optimal(
                df,
                treatment_col=self.treatment_col,
                ps_col=self.ps_col,
                pid_col="index",
                **self.kwargs,
            )
        else:
            matched = match_eager(
                df,
                treatment_col=self.treatment_col,
                ps_col=self.ps_col,
                pid_col="index",
            )
        if self.effect_type in ["ATE", "ARR"]:
            warnings.warn(
                "This is strictly speaking not ATE if we used a caliper or other matching methods. But can be interpreted as such."
            )
            return compute_matching_ate(Y, matched)
        else:
            raise ValueError(f"Effect type '{self.effect_type}' is not supported.")



# CausalEstimate/estimators/multi_estimator.py

import numpy as np
from typing import List, Dict
import pandas as pd
from CausalEstimate.core.bootstrap import generate_bootstrap_samples
from CausalEstimate.filter.propensity import filter_common_support
from CausalEstimate.estimators.base import BaseEstimator
from CausalEstimate.core.logging import log_table_stats


class MultiEstimator:
    def __init__(self, estimators: List[BaseEstimator], verbose: bool = False):
        """
        `estimators` is a list of estimator instances (AIPW, TMLE, IPW, etc.).
        Each is already configured with its own column names and effect_type.
        """
        self.estimators = estimators
        self.verbose = verbose

    def compute_effects(
        self,
        df: pd.DataFrame,
        bootstrap: bool = False,
        n_bootstraps: int = 1,
        apply_common_support: bool = False,
        common_support_threshold: float = 0.05,
    ) -> Dict[str, Dict]:
        """
        Loops over self.estimators, applies optional common support,
        optional bootstrap, and returns { "AIPW": {...}, "TMLE": {...}, ... }.
        """
        if apply_common_support:
            # Verify all estimators use the same ps_col and treatment_col
            first_estimator = self.estimators[0]
            ps_col = first_estimator.ps_col
            treatment_col = first_estimator.treatment_col
            outcome_col = first_estimator.outcome_col

            for estimator in self.estimators[1:]:
                if (
                    estimator.ps_col != ps_col
                    or estimator.treatment_col != treatment_col
                ):
                    raise ValueError(
                        "All estimators must use the same ps_col and treatment_col "
                        f"but found ps_col={estimator.ps_col} vs {ps_col} and "
                        f"treatment_col={estimator.treatment_col} vs {treatment_col}"
                    )

            df = filter_common_support(
                df,
                ps_col=ps_col,
                treatment_col=treatment_col,
                threshold=common_support_threshold,
            ).reset_index(drop=True)

        if self.verbose:
            log_table_stats(df, treatment_col, outcome_col, ps_col)

        results = {}
        for estimator in self.estimators:
            est_name = estimator.__class__.__name__

            if bootstrap and n_bootstraps > 1:
                effect, std_err = self._compute_bootstrap(estimator, df, n_bootstraps)
                results[est_name] = {
                    "effect": effect,
                    "std_err": std_err,
                    "bootstrap": True,
                    "n_bootstraps": n_bootstraps,
                }
            else:
                effect_val = estimator.compute_effect(df)
                results[est_name] = {
                    "effect": effect_val,
                    "std_err": None,
                    "bootstrap": False,
                    "n_bootstraps": 0,
                }
        return results

    def _compute_bootstrap(
        self, estimator: BaseEstimator, df: pd.DataFrame, n_bootstraps: int
    ):
        effects = []
        samples = generate_bootstrap_samples(df, n_bootstraps)
        for sample in samples:
            val = estimator.compute_effect(sample)
            effects.append(val)
        return float(np.mean(effects)), float(np.std(effects))



# test_multi_estimator.py

import unittest
import pandas as pd
import numpy as np

from CausalEstimate import MultiEstimator
from CausalEstimate.estimators.aipw import AIPW
from CausalEstimate.estimators.tmle import TMLE
from CausalEstimate.estimators.ipw import IPW
from CausalEstimate.utils.constants import (
    OUTCOME_COL,
    PS_COL,
    TREATMENT_COL,
    PROBAS_COL,
    PROBAS_T0_COL,
    PROBAS_T1_COL,
)


class TestMultiEstimator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Generate sample data once for all tests
        np.random.seed(42)
        size = 500
        epsilon = 1e-3  # Small value to avoid exact 0 or 1
        propensity_score = np.random.uniform(epsilon, 1 - epsilon, size)
        outcome_probability = np.random.uniform(epsilon, 1 - epsilon, size)
        treatment = np.random.binomial(1, propensity_score, size)
        outcome = np.random.binomial(1, outcome_probability, size)

        outcome_treated_probability = np.zeros_like(outcome_probability)
        outcome_treated_probability[treatment == 1] = outcome_probability[
            treatment == 1
        ]
        outcome_treated_probability[treatment == 0] = np.random.uniform(
            epsilon, 1 - epsilon, size
        )[treatment == 0]

        outcome_control_probability = np.zeros_like(outcome_probability)
        outcome_control_probability[treatment == 0] = outcome_probability[
            treatment == 0
        ]
        outcome_control_probability[treatment == 1] = np.random.uniform(
            epsilon, 1 - epsilon, size
        )[treatment == 1]

        cls.sample_data = pd.DataFrame(
            {
                TREATMENT_COL: treatment,
                OUTCOME_COL: outcome,
                PS_COL: propensity_score,
                PROBAS_COL: outcome_probability,
                PROBAS_T1_COL: outcome_treated_probability,
                PROBAS_T0_COL: outcome_control_probability,
            }
        )

    def _make_aipw(self):
        """
        Helper to instantiate an AIPW estimator for testing.
        """
        return AIPW(
            treatment_col=TREATMENT_COL,
            outcome_col=OUTCOME_COL,
            ps_col=PS_COL,
            probas_t1_col=PROBAS_T1_COL,
            probas_t0_col=PROBAS_T0_COL,
            effect_type="ATE",
        )

    def _make_tmle(self):
        """
        Helper to instantiate a TMLE estimator for testing.
        """
        return TMLE(
            treatment_col=TREATMENT_COL,
            outcome_col=OUTCOME_COL,
            ps_col=PS_COL,
            probas_col=PROBAS_COL,
            probas_t1_col=PROBAS_T1_COL,
            probas_t0_col=PROBAS_T0_COL,
            effect_type="ATE",
        )

    def _make_ipw(self):
        """
        Helper to instantiate an IPW estimator for testing.
        """
        return IPW(
            treatment_col=TREATMENT_COL,
            outcome_col=OUTCOME_COL,
            ps_col=PS_COL,
            effect_type="ATE",
        )

    def test_compute_effect_no_bootstrap(self):
        aipw = self._make_aipw()
        tmle = self._make_tmle()
        multi_estimator = MultiEstimator([aipw, tmle])

        # No bootstrap, no common support
        results = multi_estimator.compute_effects(
            df=self.sample_data,
            bootstrap=False,
            n_bootstraps=1,
            apply_common_support=False,
        )

        # Check that results are returned for both estimators
        self.assertIn("AIPW", results)
        self.assertIn("TMLE", results)

        # Effect estimates should be floats
        self.assertIsInstance(results["AIPW"]["effect"], float)
        self.assertIsInstance(results["TMLE"]["effect"], float)

        # No std_err when bootstrap=False
        self.assertIsNone(results["AIPW"]["std_err"])
        self.assertIsNone(results["TMLE"]["std_err"])

        # Check that bootstrap flag is False
        self.assertFalse(results["AIPW"]["bootstrap"])
        self.assertFalse(results["TMLE"]["bootstrap"])

    def test_compute_effect_with_bootstrap(self):
        aipw = self._make_aipw()
        tmle = self._make_tmle()
        multi_estimator = MultiEstimator([aipw, tmle])

        results = multi_estimator.compute_effects(
            df=self.sample_data,
            bootstrap=True,
            n_bootstraps=10,
            apply_common_support=False,
        )

        self.assertIn("AIPW", results)
        self.assertIn("TMLE", results)
        self.assertIsInstance(results["AIPW"]["effect"], float)
        self.assertIsInstance(results["TMLE"]["effect"], float)
        self.assertIsInstance(results["AIPW"]["std_err"], float)
        self.assertIsInstance(results["TMLE"]["std_err"], float)

        self.assertTrue(results["AIPW"]["bootstrap"])
        self.assertTrue(results["TMLE"]["bootstrap"])
        self.assertEqual(results["AIPW"]["n_bootstraps"], 10)
        self.assertEqual(results["TMLE"]["n_bootstraps"], 10)

    def test_missing_columns(self):
        # Here we just test that if columns are missing, the estimator complains.
        # For example, remove 'treatment' column:
        data_missing = self.sample_data.drop(columns=["treatment"])
        aipw = self._make_aipw()
        multi_estimator = MultiEstimator([aipw])

        with self.assertRaises(ValueError):
            multi_estimator.compute_effects(data_missing)

    def test_input_validation(self):
        # Introduce NaNs in the outcome column
        data_with_nan = self.sample_data.copy()
        data_with_nan.loc[0, "outcome"] = np.nan

        aipw = self._make_aipw()
        multi_estimator = MultiEstimator([aipw])

        with self.assertRaises(ValueError):
            multi_estimator.compute_effects(data_with_nan)

    def test_compute_effect_ipw(self):
        # IPW only needs treatment/outcome/ps
        ipw = self._make_ipw()
        multi_estimator = MultiEstimator([ipw])

        results = multi_estimator.compute_effects(self.sample_data)
        self.assertIn("IPW", results)
        self.assertIsInstance(results["IPW"]["effect"], float)

    def test_common_support_filtering(self):
        aipw = self._make_aipw()
        multi_estimator = MultiEstimator([aipw])

        results = multi_estimator.compute_effects(
            self.sample_data,
            apply_common_support=True,
            common_support_threshold=0.01,
            bootstrap=False,
            n_bootstraps=1,
        )
        self.assertIn("AIPW", results)
        self.assertIsInstance(results["AIPW"]["effect"], float)

    def test_parallel_bootstrap(self):
        # If you have parallel code internally, you can pass e.g. n_jobs=2 as part of your bootstrap routine.
        # We'll just test that normal bootstrapping works.
        aipw = self._make_aipw()
        multi_estimator = MultiEstimator([aipw])

        results = multi_estimator.compute_effects(
            df=self.sample_data,
            bootstrap=True,
            n_bootstraps=5,
        )
        self.assertTrue(results["AIPW"]["bootstrap"])
        self.assertEqual(results["AIPW"]["n_bootstraps"], 5)
        self.assertIsInstance(results["AIPW"]["std_err"], float)

    def test_multiple_estimators_including_ipw(self):
        aipw = self._make_aipw()
        tmle = self._make_tmle()
        ipw = self._make_ipw()
        multi_estimator = MultiEstimator([aipw, tmle, ipw])

        results = multi_estimator.compute_effects(
            df=self.sample_data, bootstrap=False, n_bootstraps=1
        )
        # We expect a dict with all three
        self.assertIn("AIPW", results)
        self.assertIn("TMLE", results)
        self.assertIn("IPW", results)


if __name__ == "__main__":
    unittest.main()



# CausalEstimate/estimators/aipw.py

import pandas as pd

from CausalEstimate.estimators.base import BaseEstimator
from CausalEstimate.estimators.functional.aipw import compute_aipw_ate, compute_aipw_att
from CausalEstimate.utils.checks import check_inputs, check_required_columns


class AIPW(BaseEstimator):
    def __init__(
        self,
        effect_type: str = "ATE",
        treatment_col: str = "treatment",
        outcome_col: str = "outcome",
        ps_col: str = "ps",
        probas_t1_col: str = "probas_t1",
        probas_t0_col: str = "probas_t0",
        **kwargs,
    ):
        super().__init__(
            effect_type=effect_type,
            treatment_col=treatment_col,
            outcome_col=outcome_col,
            ps_col=ps_col,
            **kwargs,
        )
        self.probas_t1_col = probas_t1_col
        self.probas_t0_col = probas_t0_col

    def _compute_effect(self, df: pd.DataFrame) -> float:
        check_required_columns(
            df,
            [
                self.probas_t1_col,
                self.probas_t0_col,
            ],
        )
        A = df[self.treatment_col]
        Y = df[self.outcome_col]
        ps = df[self.ps_col]
        Y1_hat = df[self.probas_t1_col]
        Y0_hat = df[self.probas_t0_col]

        check_inputs(A, Y, ps, Y1_hat=Y1_hat, Y0_hat=Y0_hat)

        if self.effect_type in ["ATE", "ARR"]:
            return compute_aipw_ate(A, Y, ps, Y0_hat, Y1_hat)
        elif self.effect_type == "ATT":
            return compute_aipw_att(A, Y, ps, Y0_hat)
        else:
            raise ValueError(f"Effect type '{self.effect_type}' is not supported.")



import numpy as np
import pandas as pd

from CausalEstimate.matching.distance import (
    compute_distance_matrix,
    filter_treated_w_insufficient_controls,
)
from CausalEstimate.utils.checks import check_required_columns, check_unique_pid
from CausalEstimate.filter.filter import filter_by_column
from CausalEstimate.matching.assignment import (
    assign_controls,
    validate_control_availability,
)
from CausalEstimate.matching.helpers import check_ps_validity
from CausalEstimate.utils.constants import (
    TREATMENT_COL,
    PS_COL,
    PID_COL,
    TREATED_PID_COL,
    CONTROL_PID_COL,
    DISTANCE_COL,
)


def match_optimal(
    df: pd.DataFrame,
    n_controls: int = 1,
    caliper: float = 0.05,
    treatment_col: str = TREATMENT_COL,
    ps_col: str = PS_COL,
    pid_col: str = PID_COL,
) -> pd.DataFrame:
    """
    Matches treated individuals to control individuals based on propensity scores
    with the option to specify the number of controls per treated individual and a caliper.

    Args:
        df (pd.DataFrame): DataFrame containing treated and control individuals.
        n_controls (int): Number of controls to match for each treated individual.
        caliper (float): Maximum allowable distance (propensity score difference) for matching.
        treatment_col (str): Column name indicating treatment status.
        ps_col (str): Column name for propensity score.
        pid_col (str): Column name for individual ID.

    Returns:
        pd.DataFrame: DataFrame with treated_pid, control_pid and distance columns.
    """
    check_required_columns(df, [treatment_col, ps_col, pid_col])
    check_unique_pid(df, pid_col)
    check_ps_validity(df, ps_col)

    treated_df = filter_by_column(df, treatment_col, 1)
    control_df = filter_by_column(df, treatment_col, 0)

    distance_matrix = compute_distance_matrix(treated_df, control_df, ps_col)

    if caliper is not None:
        distance_matrix[distance_matrix > caliper] = (
            0  # this will ignore all distances greater than the caliper
        )

    distance_matrix, treated_df = filter_treated_w_insufficient_controls(
        distance_matrix, treated_df, n_controls
    )
    validate_control_availability(treated_df, control_df, n_controls)
    # print(dist_mat)
    distance_matrix = np.repeat(
        distance_matrix, repeats=n_controls, axis=0
    )  # repeat the matrix n_controls times
    row_ind, col_ind = assign_controls(distance_matrix)

    matched_distances = distance_matrix[row_ind, col_ind].reshape(
        -1, n_controls
    )  # n_cases x n_controls
    col_ind = col_ind.reshape(-1, n_controls)  # n_cases x n_controls

    result = create_matched_df(
        matched_distances, treated_df, control_df, pid_col, n_controls, col_ind
    )
    return result


def match_eager(
    df: pd.DataFrame,
    treatment_col: str = TREATMENT_COL,
    ps_col: str = PS_COL,
    pid_col: str = PID_COL,
    caliper: float = None,
    n_controls: int = 1,
    strict: bool = True,
) -> pd.DataFrame:
    """
    Performs a greedy nearest-neighbor matching based on propensity scores,
    allowing multiple controls per treated subject. By default, n_controls=1.

    Matching proceeds in multiple "passes":
    - Pass 1: each treated tries to find its first best control
    - Pass 2: each treated tries to find its second best control, etc.

    Args:
        df (pd.DataFrame): Input dataframe.
        treatment_col (str): Name of treatment column (1=treated, 0=control).
        ps_col (str): Name of propensity score column.
        pid_col (str): Name of patient ID column.
        caliper (float, optional): Maximum allowed absolute difference in PS for matching.
                                   If no control is within the caliper, that treated subject
                                   remains unmatched (or raises ValueError if strict=True).
        n_controls (int): How many distinct control matches to find per treated subject.
        strict (bool): If True, raise a ValueError if any treated subject fails to find
                       a control at any pass. If False, skip unmatched passes.

    Returns:
        pd.DataFrame with columns [treated_pid, control_pid, distance].
        This may contain up to (n_controls * number_of_treated) rows,
        if all can be matched on every pass.

    Raises:
        ValueError: If strict=True and a treated subject cannot be matched on any pass.
    """
    # Separate treated vs. control
    treated_array = df.loc[df[treatment_col] == 1, [pid_col, ps_col]].values
    control_array = df.loc[df[treatment_col] == 0, [pid_col, ps_col]].values

    # Keep track of how many matches each treated has found
    # Map from treated_pid -> current match count
    # (only needed if we want to track partial matches in strict mode)
    match_count = {t_pid: 0 for t_pid, _ in treated_array}

    all_matches = []
    used_control = set()  # keep track of which controls are already matched

    # Repeat the greedy pass n_controls times
    for pass_index in range(n_controls):
        pass_matches = []  # store matches found in this pass

        for t_pid, t_ps in treated_array:
            # If this treated subject is already fully matched in previous passes, skip
            if match_count[t_pid] >= pass_index + 1:
                # already got pass_index matches
                continue

            # compute distance to all controls
            ps_diffs = np.abs(control_array[:, 1] - t_ps)

            # apply caliper if specified
            if caliper is not None:
                within_caliper = ps_diffs <= caliper
                if not within_caliper.any():
                    if strict:
                        raise ValueError(
                            f"Treated subject {t_pid} cannot find a control within caliper={caliper} "
                            f"on pass {pass_index+1} of {n_controls}."
                        )
                    else:
                        continue
                ps_diffs = ps_diffs[within_caliper]
                valid_control = control_array[within_caliper]
            else:
                valid_control = control_array

            if len(valid_control) == 0:
                if strict:
                    raise ValueError(
                        f"Treated subject {t_pid} cannot find any available control on pass {pass_index+1}. "
                        f"All possible controls are used or out of range."
                    )
                else:
                    continue

            # sort by smallest distance
            sorted_indices = np.argsort(ps_diffs)
            found_match = False

            for idx in sorted_indices:
                c_pid = valid_control[idx, 0]
                c_ps = valid_control[idx, 1]
                if c_pid not in used_control:
                    dist = abs(t_ps - c_ps)
                    pass_matches.append([t_pid, c_pid, dist])
                    used_control.add(c_pid)  # can't use it again
                    match_count[t_pid] += 1
                    found_match = True
                    break

            if not found_match and strict:
                raise ValueError(
                    f"Treated subject {t_pid} cannot find an available control on pass {pass_index+1}, "
                    "either because all valid controls are used up or out of range."
                )

        # merge pass_matches into our global list
        all_matches.extend(pass_matches)

    return pd.DataFrame(
        all_matches, columns=[TREATED_PID_COL, CONTROL_PID_COL, DISTANCE_COL]
    )


def create_matched_df(
    matched_distances: np.array,
    treated_df: pd.DataFrame,
    control_df: pd.DataFrame,
    pid_col: str,
    n_controls: int,
    col_ind: np.array,
) -> pd.DataFrame:
    """
    Creates a DataFrame of matched treated-control pairs and their distances.
    """
    treated_ids_repeated = np.repeat(treated_df[pid_col].values, n_controls)
    control_ids = control_df.iloc[col_ind.flatten()][pid_col].values
    return pd.DataFrame(
        {
            TREATED_PID_COL: treated_ids_repeated,
            CONTROL_PID_COL: control_ids,
            DISTANCE_COL: matched_distances.flatten(),
        }
    )



PID_COL = "PID"
TREATMENT_COL = "treatment"
OUTCOME_COL = "outcome"
PS_COL = "ps"
PROBAS_T1_COL = "probas_t1"  # probability of outcome when treated
PROBAS_T0_COL = "probas_t0"  # probability of outcome when not treated
PROBAS_COL = "probas"  # probability of outcome under given treatment


OUTCOME_COL = "Y"
OUTCOME_CF_COL = "Y_cf"

TREATED_PID_COL = "treated_pid_col"
CONTROL_PID_COL = "control_pid_col"
DISTANCE_COL = "distance_col"



import warnings

import numpy as np
from scipy.special import expit, logit
from statsmodels.genmod.families import Binomial
from statsmodels.genmod.generalized_linear_model import GLM


def compute_tmle_ate(A, Y, ps, Y0_hat, Y1_hat, Yhat):
    """
    Estimate the average treatment effect using the targeted maximum likelihood estimation (TMLE) method.

    Parameters:
    -----------
    A: array-like
        Treatment assignment (0 or 1)
    Y: array-like
        Binary outcome
    ps: array-like
        Propensity score P(A=1|X)
    Y0_hat: array-like
        Initial outcome prediction for control group P(Y|A=0,X)
    Y1_hat: array-like
        Initial outcome prediction for treatment group P(Y|A=1,X)
    Yhat: array-like
        Combined outcome prediction (A*Y1_hat + (1-A)*Y0_hat)

    Returns:
    --------
    float: Average treatment effect estimate
    """
    Q_star_1, Q_star_0 = compute_estimates(A, Y, ps, Y0_hat, Y1_hat, Yhat)
    return (Q_star_1 - Q_star_0).mean()


def compute_tmle_rr(A, Y, ps, Y0_hat, Y1_hat, Yhat):
    """
    Estimate the risk ratio using the targeted maximum likelihood estimation (TMLE) method.

    Parameters:
    -----------
    A: array-like
        Treatment assignment (0 or 1)
    Y: array-like
        Binary outcome
    ps: array-like
        Propensity score P(A=1|X)
    Y0_hat: array-like
        Initial outcome prediction for control group P(Y|A=0,X)
    Y1_hat: array-like
        Initial outcome prediction for treatment group P(Y|A=1,X)
    Yhat: array-like
        Combined outcome prediction (A*Y1_hat + (1-A)*Y0_hat)

    Returns:
    --------
    float: Risk ratio estimate
    """
    Q_star_1, Q_star_0 = compute_estimates(A, Y, ps, Y0_hat, Y1_hat, Yhat)
    Q_star_0_m = Q_star_0.mean()
    if Q_star_0_m == 0:
        warnings.warn("Q_star_0 is 0, returning inf", RuntimeWarning)
        return np.inf
    return Q_star_1.mean() / Q_star_0_m


def compute_estimates(A, Y, ps, Y0_hat, Y1_hat, Yhat):
    """
    Compute updated outcome estimates using TMLE targeting step.

    Parameters:
    -----------
    A: array-like
        Treatment assignment (0 or 1)
    Y: array-like
        Binary outcome
    ps: array-like
        Propensity score P(A=1|X)
    Y0_hat: array-like
        Initial outcome prediction for control group P(Y|A=0,X)
    Y1_hat: array-like
        Initial outcome prediction for treatment group P(Y|A=1,X)
    Yhat: array-like
        Combined outcome prediction (A*Y1_hat + (1-A)*Y0_hat)

    Returns:
    --------
    tuple: (Q_star_1, Q_star_0)
        Updated outcome predictions for treatment and control groups
    """
    epsilon = estimate_fluctuation_parameter(A, Y, ps, Yhat)
    Q_star_1, Q_star_0 = update_estimates(ps, Y0_hat, Y1_hat, epsilon)
    return Q_star_1, Q_star_0


def update_estimates(ps, Y0_hat, Y1_hat, epsilon):
    """
    Update the initial outcome estimates using the fluctuation parameter.

    Parameters:
    -----------
    ps: array-like
        Propensity score
    Y0_hat: array-like
        Initial outcome prediction for control group
    Y1_hat: array-like
        Initial outcome prediction for treatment group
    epsilon: float
        Estimated fluctuation parameter

    Returns:
    --------
    tuple: (Q_star_1, Q_star_0)
        Updated outcome predictions for treatment and control groups
    """
    # Define clever covariates
    H1 = 1.0 / ps
    H0 = -1.0 / (1.0 - ps)

    # Update initial estimates with targeting step
    Q_star_1 = expit(logit(Y1_hat) + epsilon * H1)
    Q_star_0 = expit(logit(Y0_hat) + epsilon * H0)

    return Q_star_1, Q_star_0


def estimate_fluctuation_parameter(A, Y, ps, Yhat) -> float:
    """
    Estimate the fluctuation parameter epsilon using a logistic regression model.
    Returns the estimated epsilon.
    """
    # compute the clever covariate H
    H = A / ps - (1 - A) / (1 - ps)
    # Check for extreme values in clever covariate
    if np.any(np.abs(H) > 100):
        warnings.warn(
            "Extreme values detected in clever covariate H. "
            "This may indicate issues with propensity scores near 0 or 1.",
            RuntimeWarning,
        )
    # Use logit of the current outcome as offset
    offset = logit(Yhat)

    # Fit the model with offset
    model = GLM(Y, H, family=Binomial(), offset=offset).fit()
    return np.asarray(model.params)[0]
